#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>

#define POST_SCALER 0xFF

typedef struct {
    uint8_t duty_cycle; // Expressed as a percentage (0 - 100)
    int8_t direction; // Either 1 or -1
    uint16_t post_scale; // Used as a software postscaler
} PWM_Data;

// Use volatile to prevent compilier optimization on this global variable
volatile PWM_Data g_pwm = {0};

// Convert a % duty cycle to an 8-bit value for the compare register
void set_duty(uint8_t percentage){
    // Bound checking 
    if(percentage > 100){
        percentage = 100;
    }

    // Calculate the percent value of OCR0A
    uint16_t temp = (uint16_t) OCR0A * (uint16_t) percentage;
    OCR0B = (uint8_t) (temp / 100);
}

ISR(PCINT0_vect){
    // Handle the interrupt here (nothing to do...)
}

// Timer0 overflow interrupt
ISR(TIM0_OVF_vect){
    // Handle the software postscaler
    g_pwm.post_scale++;
    if(g_pwm.post_scale < POST_SCALER){
        return;
    }
    g_pwm.post_scale = 0;

    // Start interrupt
    int16_t new_duty = g_pwm.duty_cycle + (g_pwm.direction * (g_pwm.duty_cycle / 32 + 1));

    // Check bounds
    if(new_duty <= 0){
        new_duty = 0;
        g_pwm.direction = 1;
    } else if (new_duty >= 100){
        new_duty = 100;
        g_pwm.direction = -1;
    }
    g_pwm.duty_cycle = (uint8_t) new_duty;

    set_duty(g_pwm.duty_cycle);
}


// Use this timer for PWM
void init_timer0(){
    // Non-inverting PWM mode on OC0B
    TCCR0A = (1 << COM0B1);

    // Fast PWM mode
    TCCR0A |= (1 << WGM02) | (1 << WGM01) | (1 << WGM00);

    OCR0A = 0xFF;

    // 50% duty cycle at startup
    OCR0B = 0xFF / 2;

    // Set PWM pin as an output
    DDRB |= (1 << PB1);

    // Enable Timer overflow interrupt
    TIMSK |= (1 << TOIE0);

    // Start timer with no prescaler
    TCCR0B = (1 << CS00);
}

void init_PWM(){
    // Setup default settings
    g_pwm.duty_cycle = 50;
    g_pwm.direction = 1;

    init_timer0();

    set_duty(g_pwm.duty_cycle);
}

void sleep(){
    TCCR0A &= ~((1 << COM0B1) | (1 << COM0B0)); // Disconnect OC0B from PB1 to turn it off
    PORTB &= ~(1 << PB1);

    set_sleep_mode(SLEEP_MODE_PWR_DOWN);
    sleep_enable(); // This doesn't start sleep, but it allows up to go to sleep using sleep_cpu()

    sleep_cpu(); // Put the system to sleep

    sleep_disable(); // Wake up here

    TCCR0A |= (1 << COM0B1); // Re-enable OC0B control over PB1
}

void init_button(){
    // Make our button an input
    DDRB &= ~(1 << PB2);

    // Pin change interrupt on PB2
    GIMSK |= (1 << PCIE); // Enable pin change interrupts
    PCMSK |= (1 << PCINT2); // Use PB2 as our interrupt source
}

int main(){
    // Init PWM on PB1
    init_PWM();

    init_button();

    sei();

    while(1){
        _delay_ms(1000 * 10); // 10 seconds delay
        sleep();
    }
}
